==> xpkg-v1.0.16.tar.bz2: xpkg-v1.0.16/httpcli/httpcli.go <==
// Package httpcli is http request client, which only supports return json format.
package httpcli

import (
	"bytes"
	"context"
	"encoding/json"
	"errors"
	"fmt"
	"io"
	"net/http"
	"net/url"
	"strings"
	"time"

	"gitlab.wanzhuangkj.com/tush/xpkg/gin/xctx"
	"gitlab.wanzhuangkj.com/tush/xpkg/logger"
	"gitlab.wanzhuangkj.com/tush/xpkg/utils/xjson"
	"gitlab.wanzhuangkj.com/tush/xpkg/xerrors/xerror"

	"github.com/duke-git/lancet/v2/retry"
)

const defaultTimeout = 30 * time.Second

var (
	contentMark = []byte(" ...... ")
)

// Request HTTP request
type Request struct {
	customRequest func(req *http.Request, data *bytes.Buffer) // used to define HEADER, e.g. to add sign, etc.
	url           string
	params        map[string]interface{} // parameters after URL
	body          string                 // Body data
	bodyJSON      interface{}            // JSON marshal body data
	timeout       time.Duration          // Client timeout
	retryCount    uint
	headers       map[string]string

	request  *http.Request
	response *Response
	method   string
	err      error
}

// Response HTTP response
type Response struct {
	Body       io.ReadCloser
	StatusCode int
	Status     string // e.g. "200 OK"
	Header     http.Header
	err        error
}

// -----------------------------------  Request way 1 -----------------------------------

// New create a new Request
func New() *Request {
	return &Request{}
}

// Reset set all fields to default value, use at pool
func (req *Request) Reset() {
	req.params = nil
	req.body = ""
	req.bodyJSON = nil
	req.timeout = 0
	req.headers = nil

	req.request = nil
	req.response = nil
	req.method = ""
	req.err = nil
}

// SetURL set URL
func (req *Request) SetURL(path string) *Request {
	req.url = path
	return req
}

// SetParams parameters after setting the URL
func (req *Request) SetParams(params map[string]interface{}) *Request {
	if req.params == nil {
		req.params = params
	} else {
		for k, v := range params {
			req.params[k] = v
		}
	}
	return req
}

// SetParam parameters after setting the URL
func (req *Request) SetParam(k string, v interface{}) *Request {
	if req.params == nil {
		req.params = make(map[string]interface{})
	}
	req.params[k] = v
	return req
}

// SetBody set body data, support string and []byte, if it is not string, it will be json marshal.
func (req *Request) SetBody(body interface{}) *Request {
	switch v := body.(type) {
	case string:
		req.body = v
	case []byte:
		req.body = string(v)
	default:
		req.bodyJSON = body
	}
	return req
}

// SetTimeout set timeout
func (req *Request) SetTimeout(t time.Duration) *Request {
	req.timeout = t
	return req
}
func (req *Request) SetRetry(count uint) *Request {
	req.retryCount = count
	return req
}

// SetContentType set ContentType
func (req *Request) SetContentType(a string) *Request {
	req.SetHeader("Content-Type", a)
	return req
}

// SetHeader set the value of the request header
func (req *Request) SetHeader(k, v string) *Request {
	if req.headers == nil {
		req.headers = make(map[string]string)
	}
	req.headers[k] = v
	return req
}

// SetHeaders set the value of Request Headers
func (req *Request) SetHeaders(headers map[string]string) *Request {
	if req.headers == nil {
		req.headers = make(map[string]string)
	}
	for k, v := range headers {
		req.headers[k] = v
	}
	return req
}

// CustomRequest customize request, e.g. add sign, set header, etc.
func (req *Request) CustomRequest(f func(req *http.Request, data *bytes.Buffer)) *Request {
	req.customRequest = f
	return req
}

// GET send a GET request
func (req *Request) GET(ctx context.Context) (*Response, error) {
	req.method = http.MethodGet
	return req.pull(ctx)
}

// DELETE send a DELETE request
func (req *Request) DELETE(ctx context.Context) (*Response, error) {
	req.method = http.MethodDelete
	return req.pull(ctx)
}

// POST send a POST request
func (req *Request) POST(ctx context.Context) (*Response, error) {
	req.method = http.MethodPost
	return req.push(ctx)
}

// PUT send a PUT request
func (req *Request) PUT(ctx context.Context) (*Response, error) {
	req.method = http.MethodPut
	return req.push(ctx)
}

// PATCH send PATCH requests
func (req *Request) PATCH(ctx context.Context) (*Response, error) {
	req.method = http.MethodPatch
	return req.push(ctx)
}

// Do a request
func (req *Request) Do(ctx context.Context, method string, data interface{}) (*Response, error) {
	req.method = method

	switch method {
	case http.MethodGet, http.MethodDelete:
		if data != nil {
			if params, ok := data.(map[string]interface{}); ok { //nolint
				req.SetParams(params)
			} else {
				req.err = errors.New("params is not a map[string]interface{}")
				return nil, req.err
			}
		}

		return req.pull(ctx)

	case http.MethodPost, http.MethodPut, http.MethodPatch:
		if data != nil {
			req.SetBody(data)
		}

		return req.push(ctx)
	}

	req.err = errors.New("unknow method " + method)
	return nil, req.err
}

func (req *Request) pull(ctx context.Context) (*Response, error) {
	val := ""
	if len(req.params) > 0 {
		values := url.Values{}
		for k, v := range req.params {
			values.Add(k, fmt.Sprintf("%v", v))
		}
		val += values.Encode()
	}

	if val != "" {
		if strings.Contains(req.url, "?") {
			req.url += "&" + val
		} else {
			req.url += "?" + val
		}
	}

	var buf *bytes.Buffer
	if req.customRequest != nil {
		buf = bytes.NewBufferString(val)
	}

	return req.send(ctx, nil, buf)
}

func (req *Request) push(ctx context.Context) (*Response, error) {
	var buf *bytes.Buffer

	if req.bodyJSON != nil {
		body, err := json.Marshal(req.bodyJSON)
		if err != nil {
			req.err = err
			return nil, req.err
		}
		buf = bytes.NewBuffer(body)
	} else {
		buf = bytes.NewBufferString(req.body)
	}

	logger.Info("httpcli req", logger.String("body", buf.String()), xctx.CtxTraceIDField(ctx))
	return req.send(ctx, buf, buf)
}

func (req *Request) ReqString(ctx context.Context) string {
	request := make(map[string]any)
	request["method"] = req.request.Method
	request["url"] = req.request.URL.String()
	request["headers"] = req.request.Header
	if req.request.Method == http.MethodPost || req.request.Method == http.MethodPut || req.request.Method == http.MethodPatch || req.request.Method == http.MethodDelete {
		if req.request.Body != nil {
			buf := bytes.Buffer{}
			n, _ := buf.ReadFrom(req.request.Body)
			request["size"] = buf.Len()
			if n > 0 {
				request["body"] = getRequestBody(&buf, 256)
			}
		}
	}
	return xjson.ToJsonString(request)
}

func getRequestBody(buf *bytes.Buffer, maxLen int) []byte {
	l := buf.Len()
	if l == 0 {
		return []byte("")
	} else if l < maxLen {
		return buf.Bytes()
	}

	body := make([]byte, maxLen)
	copy(body, buf.Bytes())
	return append(body[:maxLen-len(contentMark)], contentMark...)
}

func (req *Request) send(ctx context.Context, body io.Reader, buf *bytes.Buffer) (*Response, error) {
	logger.Info("httpcli req", logger.String("url", req.url), xctx.CtxTraceIDField(ctx))
	req.request, req.err = http.NewRequest(req.method, req.url, body)
	if req.err != nil {
		return nil, req.err
	}
	if buf != nil {
		requestID := xctx.CtxRequestID(ctx)
		xctx.SetVal(ctx, fmt.Sprintf("oApiReq-%s", requestID), buf.String())
	}
	if req.customRequest != nil {
		req.customRequest(req.request, buf)
	}

	if req.headers != nil {
		for k, v := range req.headers {
			req.request.Header.Add(k, v)
		}
	}

	req.request.Header.Add(xctx.HeaderXRequestIDKey, xctx.CtxRequestID(ctx))

	if req.timeout < 1 {
		req.timeout = defaultTimeout
	}

	client := http.Client{Timeout: req.timeout}
	resp := new(Response)
	resp.Header = make(http.Header)

	ctx, cancel := context.WithTimeout(ctx, time.Second*20)
	defer cancel()
	var err error
	if req.retryCount > 0 {
		err = retry.Retry(func() error {
			response, err := client.Do(req.request)
			if err != nil {
				logger.Info("httpcli fail", logger.Any("err", err), xctx.CtxTraceIDField(ctx))
				return xerror.Wrap(err, "client do")
			}
			if response != nil {
				resp.Status = response.Status
				resp.StatusCode = response.StatusCode
				for k, v := range response.Header {
					resp.Header[k] = v
				}
				defer response.Body.Close()
				body, err := io.ReadAll(response.Body)
				if err != nil {
					return xerror.New(err.Error())
				}
				bodyBuf := bytes.NewBuffer(body)
				xctx.SetVal(ctx, fmt.Sprintf("oApiRsp-%s", xctx.CtxRequestID(ctx)), buf.String())
				logger.Info("httpcli rsp", logger.Any("body", bodyBuf.String()), xctx.CtxTraceIDField(ctx))
				if body != nil {
					resp.Body = io.NopCloser(bodyBuf)
				}
			}
			return nil
		},
			retry.RetryTimes(req.retryCount),
			retry.RetryWithLinearBackoff(2*time.Second),
			retry.Context(ctx),
		)
	} else {
		response, e := client.Do(req.request)
		if e != nil {
			err = xerror.New(e.Error())
		}
		if response != nil {
			resp.Status = response.Status
			resp.StatusCode = response.StatusCode
			for k, v := range response.Header {
				resp.Header[k] = v
			}
			defer response.Body.Close()
			body, e := io.ReadAll(response.Body)
			if e != nil {
				err = xerror.New(e.Error())
			}
			bodyBuf := bytes.NewBuffer(body)
			xctx.SetVal(ctx, fmt.Sprintf("oApiRsp-%s", xctx.CtxRequestID(ctx)), buf.String())
			logger.Info("httpcli rsp", logger.Any("body", bodyBuf.String()), xctx.CtxTraceIDField(ctx))
			if body != nil {
				resp.Body = io.NopCloser(bodyBuf)
			}
		}
	}

	if err != nil {
		logger.Error("httpcli fail", logger.Any("err", err), xctx.CtxTraceIDField(ctx))
	}

	req.response = resp
	req.err = err
	resp.err = err

	return resp, resp.err
}

// Response return response
func (req *Request) Response() (*Response, error) {
	if req.err != nil {
		return nil, req.err
	}
	return req.response, req.response.Error()
}

// -----------------------------------  Response -----------------------------------

// Error return err
func (resp *Response) Error() error {
	return resp.err
}

// BodyString returns the body data of the HttpResponse
func (resp *Response) BodyString() (string, error) {
	if resp.err != nil {
		return "", resp.err
	}
	body, err := resp.ReadBody()
	return string(body), err
}

// BodyString returns the body data of the HttpResponse
func (resp *Response) RspString() string {
	ret := make(map[string]any)
	ret["status"] = resp.Status
	ret["statusCode"] = resp.StatusCode
	ret["header"] = resp.Header
	if resp.err != nil {
		ret["err"] = resp.err.Error()
	}
	body, err := resp.ReadBody()
	if err == nil {
		ret["body"] = bytes.NewBuffer(body).String()
	}
	return xjson.ToJsonString(ret)
}

// ReadBody returns the body data of the HttpResponse
func (resp *Response) ReadBody() ([]byte, error) {
	if resp.err != nil {
		return []byte{}, resp.err
	}

	if resp.Body == nil {
		return []byte{}, errors.New("nil")
	}

	body, err := io.ReadAll(resp.Body)
	if err != nil {
		return []byte{}, err
	}
	resp.Body.Close()

	resp.Body = io.NopCloser(bytes.NewBuffer(body))
	return body, nil
}

// BindJSON parses the response's body as JSON
func (resp *Response) BindJSON(v interface{}) error {
	if resp.err != nil {
		return resp.err
	}
	body, err := resp.ReadBody()
	if err != nil {
		return err
	}
	return json.Unmarshal(body, v)
}

// -----------------------------------  Request way 2 -----------------------------------

// Option set options.
type Option func(*options)

type options struct {
	params  map[string]interface{}
	headers map[string]string
	timeout time.Duration
}

func (o *options) apply(opts ...Option) {
	for _, opt := range opts {
		opt(o)
	}
}

func defaultOptions() *options {
	return &options{}
}

// WithParams set params
func WithParams(params map[string]interface{}) Option {
	return func(o *options) {
		o.params = params
	}
}

// WithHeaders set headers
func WithHeaders(headers map[string]string) Option {
	return func(o *options) {
		o.headers = headers
	}
}

// WithTimeout set timeout
func WithTimeout(t time.Duration) Option {
	return func(o *options) {
		o.timeout = t
	}
}

// Get request, return custom json format
func Get(ctx context.Context, result interface{}, urlStr string, opts ...Option) error {
	o := defaultOptions()
	o.apply(opts...)
	return gDo(ctx, "GET", result, urlStr, o.params, o.headers, o.timeout)
}

// Delete request, return custom json format
func Delete(ctx context.Context, result interface{}, urlStr string, opts ...Option) error {
	o := defaultOptions()
	o.apply(opts...)
	return gDo(ctx, "DELETE", result, urlStr, o.params, o.headers, o.timeout)
}

// Post request, return custom json format
func Post(ctx context.Context, result interface{}, urlStr string, body interface{}, opts ...Option) error {
	o := defaultOptions()
	o.apply(opts...)
	return do(ctx, "POST", result, urlStr, body, o.params, o.headers, o.timeout)
}

// Put request, return custom json format
func Put(ctx context.Context, result interface{}, urlStr string, body interface{}, opts ...Option) error {
	o := defaultOptions()
	o.apply(opts...)
	return do(ctx, "PUT", result, urlStr, body, o.params, o.headers, o.timeout)
}

// Patch request, return custom json format
func Patch(ctx context.Context, result interface{}, urlStr string, body interface{}, opts ...Option) error {
	o := defaultOptions()
	o.apply(opts...)
	return do(ctx, "PATCH", result, urlStr, body, o.params, o.headers, o.timeout)
}

var requestErr = func(err error) error { return fmt.Errorf("request error, err=%v", err) }
var jsonParseErr = func(err error) error { return fmt.Errorf("json parsing error, err=%v", err) }
var notOKErr = func(resp *Response) error {
	body, err := resp.ReadBody()
	if err != nil {
		return err
	}
	if len(body) > 500 {
		body = append(body[:500], []byte(" ......")...)
	}
	return fmt.Errorf("statusCode=%d, body=%s", resp.StatusCode, body)
}

func do(ctx context.Context, method string, result interface{}, urlStr string, body interface{}, params KV, headers map[string]string, timeout time.Duration) error {
	if result == nil {
		return fmt.Errorf("'result' can not be nil")
	}

	req := &Request{}
	req.SetURL(urlStr)
	req.SetContentType("application/json")
	req.SetParams(params)
	req.SetHeaders(headers)
	req.SetBody(body)
	req.SetTimeout(timeout)

	var resp *Response
	var err error
	switch method {
	case "POST":
		resp, err = req.POST(ctx)
	case "PUT":
		resp, err = req.PUT(ctx)
	case "PATCH":
		resp, err = req.PATCH(ctx)
	}
	if err != nil {
		return requestErr(err)
	}
	defer resp.Body.Close() //nolint

	if resp.StatusCode != 200 {
		return notOKErr(resp)
	}

	err = resp.BindJSON(result)
	if err != nil {
		return jsonParseErr(err)
	}

	return nil
}

func gDo(ctx context.Context, method string, result interface{}, urlStr string, params KV, headers map[string]string, timeout time.Duration) error {
	req := &Request{}
	req.SetURL(urlStr)
	req.SetParams(params)
	req.SetHeaders(headers)
	req.SetTimeout(timeout)

	var resp *Response
	var err error
	switch method {
	case "GET":
		resp, err = req.GET(ctx)
	case "DELETE":
		resp, err = req.DELETE(ctx)
	}
	if err != nil {
		return requestErr(err)
	}
	defer resp.Body.Close() //nolint

	if resp.StatusCode != 200 {
		return notOKErr(resp)
	}

	err = resp.BindJSON(result)
	if err != nil {
		return jsonParseErr(err)
	}

	return nil
}

// StdResult standard return data
type StdResult struct {
	Code int         `json:"code"`
	Msg  string      `json:"msg"`
	Data interface{} `json:"data,omitempty"`
}

// StdResult standard return data
type StdResult2 struct {
	Code int         `json:"code"`
	Msg  string      `json:"message"`
	Data interface{} `json:"data,omitempty"`
}

// KV string:interface{}
type KV = map[string]interface{}

==> xpkg-v1.0.19.tar.bz2: xpkg-v1.0.19/httpcli/httpcli.go <==
// Package httpcli is http request client, which only supports return json format.
package httpcli

import (
	"bytes"
	"context"
	"encoding/json"
	"errors"
	"fmt"
	"io"
	"net/http"
	"net/url"
	"strings"
	"time"

	"gitlab.wanzhuangkj.com/tush/xpkg/gin/xctx"
	"gitlab.wanzhuangkj.com/tush/xpkg/logger"
	"gitlab.wanzhuangkj.com/tush/xpkg/utils/xjson"
	"gitlab.wanzhuangkj.com/tush/xpkg/xerrors/xerror"

	"github.com/duke-git/lancet/v2/retry"
)

const defaultTimeout = 30 * time.Second

var (
	contentMark = []byte(" ...... ")
)

// Request HTTP request
type Request struct {
	customRequest func(req *http.Request, data *bytes.Buffer) // used to define HEADER, e.g. to add sign, etc.
	url           string
	params        map[string]interface{} // parameters after URL
	body          string                 // Body data
	bodyJSON      interface{}            // JSON marshal body data
	timeout       time.Duration          // Client timeout
	retryCount    uint
	headers       map[string]string

	request  *http.Request
	response *Response
	method   string
	err      error
}

// Response HTTP response
type Response struct {
	Body       io.ReadCloser
	StatusCode int
	Status     string // e.g. "200 OK"
	Header     http.Header
	err        error
}

// -----------------------------------  Request way 1 -----------------------------------

// New create a new Request
func New() *Request {
	return &Request{}
}

// Reset set all fields to default value, use at pool
func (req *Request) Reset() {
	req.params = nil
	req.body = ""
	req.bodyJSON = nil
	req.timeout = 0
	req.headers = nil

	req.request = nil
	req.response = nil
	req.method = ""
	req.err = nil
}

// SetURL set URL
func (req *Request) SetURL(path string) *Request {
	req.url = path
	return req
}

// SetParams parameters after setting the URL
func (req *Request) SetParams(params map[string]interface{}) *Request {
	if req.params == nil {
		req.params = params
	} else {
		for k, v := range params {
			req.params[k] = v
		}
	}
	return req
}

// SetParam parameters after setting the URL
func (req *Request) SetParam(k string, v interface{}) *Request {
	if req.params == nil {
		req.params = make(map[string]interface{})
	}
	req.params[k] = v
	return req
}

// SetBody set body data, support string and []byte, if it is not string, it will be json marshal.
func (req *Request) SetBody(body interface{}) *Request {
	switch v := body.(type) {
	case string:
		req.body = v
	case []byte:
		req.body = string(v)
	default:
		req.bodyJSON = body
	}
	return req
}

// SetTimeout set timeout
func (req *Request) SetTimeout(t time.Duration) *Request {
	req.timeout = t
	return req
}
func (req *Request) SetRetry(count uint) *Request {
	req.retryCount = count
	return req
}

// SetContentType set ContentType
func (req *Request) SetContentType(a string) *Request {
	req.SetHeader("Content-Type", a)
	return req
}

// SetHeader set the value of the request header
func (req *Request) SetHeader(k, v string) *Request {
	if req.headers == nil {
		req.headers = make(map[string]string)
	}
	req.headers[k] = v
	return req
}

// SetHeaders set the value of Request Headers
func (req *Request) SetHeaders(headers map[string]string) *Request {
	if req.headers == nil {
		req.headers = make(map[string]string)
	}
	for k, v := range headers {
		req.headers[k] = v
	}
	return req
}

// CustomRequest customize request, e.g. add sign, set header, etc.
func (req *Request) CustomRequest(f func(req *http.Request, data *bytes.Buffer)) *Request {
	req.customRequest = f
	return req
}

// GET send a GET request
func (req *Request) GET(ctx context.Context) (*Response, error) {
	req.method = http.MethodGet
	return req.pull(ctx)
}

// DELETE send a DELETE request
func (req *Request) DELETE(ctx context.Context) (*Response, error) {
	req.method = http.MethodDelete
	return req.pull(ctx)
}

// POST send a POST request
func (req *Request) POST(ctx context.Context) (*Response, error) {
	req.method = http.MethodPost
	return req.push(ctx)
}

// PUT send a PUT request
func (req *Request) PUT(ctx context.Context) (*Response, error) {
	req.method = http.MethodPut
	return req.push(ctx)
}

// PATCH send PATCH requests
func (req *Request) PATCH(ctx context.Context) (*Response, error) {
	req.method = http.MethodPatch
	return req.push(ctx)
}

// Do a request
func (req *Request) Do(ctx context.Context, method string, data interface{}) (*Response, error) {
	req.method = method

	switch method {
	case http.MethodGet, http.MethodDelete:
		if data != nil {
			if params, ok := data.(map[string]interface{}); ok { //nolint
				req.SetParams(params)
			} else {
				req.err = errors.New("params is not a map[string]interface{}")
				return nil, req.err
			}
		}

		return req.pull(ctx)

	case http.MethodPost, http.MethodPut, http.MethodPatch:
		if data != nil {
			req.SetBody(data)
		}

		return req.push(ctx)
	}

	req.err = errors.New("unknow method " + method)
	return nil, req.err
}

func (req *Request) pull(ctx context.Context) (*Response, error) {
	val := ""
	if len(req.params) > 0 {
		values := url.Values{}
		for k, v := range req.params {
			values.Add(k, fmt.Sprintf("%v", v))
		}
		val += values.Encode()
	}

	if val != "" {
		if strings.Contains(req.url, "?") {
			req.url += "&" + val
		} else {
			req.url += "?" + val
		}
	}

	var buf *bytes.Buffer
	if req.customRequest != nil {
		buf = bytes.NewBufferString(val)
	}

	return req.send(ctx, nil, buf)
}

func (req *Request) push(ctx context.Context) (*Response, error) {
	var buf *bytes.Buffer

	if req.bodyJSON != nil {
		body, err := json.Marshal(req.bodyJSON)
		if err != nil {
			req.err = err
			return nil, req.err
		}
		buf = bytes.NewBuffer(body)
	} else {
		buf = bytes.NewBufferString(req.body)
	}

	logger.Info("httpcli req", logger.String("body", buf.String()), xctx.CtxTraceIDField(ctx))
	return req.send(ctx, buf, buf)
}

func (req *Request) ReqString(ctx context.Context) string {
	request := make(map[string]any)
	request["method"] = req.request.Method
	request["url"] = req.request.URL.String()
	request["headers"] = req.request.Header
	if req.request.Method == http.MethodPost || req.request.Method == http.MethodPut || req.request.Method == http.MethodPatch || req.request.Method == http.MethodDelete {
		if req.request.Body != nil {
			buf := bytes.Buffer{}
			n, _ := buf.ReadFrom(req.request.Body)
			request["size"] = buf.Len()
			if n > 0 {
				request["body"] = getRequestBody(&buf, 256)
			}
		}
	}
	return xjson.ToJsonString(request)
}

func getRequestBody(buf *bytes.Buffer, maxLen int) []byte {
	l := buf.Len()
	if l == 0 {
		return []byte("")
	} else if l < maxLen {
		return buf.Bytes()
	}

	body := make([]byte, maxLen)
	copy(body, buf.Bytes())
	return append(body[:maxLen-len(contentMark)], contentMark...)
}

func (req *Request) send(ctx context.Context, body io.Reader, buf *bytes.Buffer) (*Response, error) {
	logger.Info("httpcli req", logger.String("url", req.url), xctx.CtxTraceIDField(ctx))
	req.request, req.err = http.NewRequest(req.method, req.url, body)
	if req.err != nil {
		return nil, req.err
	}
	if buf != nil {
		requestID := xctx.CtxRequestID(ctx)
		xctx.SetVal(ctx, fmt.Sprintf("oApiReq-%s", requestID), buf.Bytes())
	}
	if req.customRequest != nil {
		req.customRequest(req.request, buf)
	}

	if req.headers != nil {
		for k, v := range req.headers {
			req.request.Header.Add(k, v)
		}
	}

	req.request.Header.Add(xctx.HeaderXRequestIDKey, xctx.CtxRequestID(ctx))

	if req.timeout < 1 {
		req.timeout = defaultTimeout
	}

	client := http.Client{Timeout: req.timeout}
	resp := new(Response)
	resp.Header = make(http.Header)

	ctx, cancel := context.WithTimeout(ctx, time.Second*20)
	defer cancel()
	var err error
	if req.retryCount > 0 {
		err = retry.Retry(func() error {
			response, err := client.Do(req.request)
			if err != nil {
				logger.Info("httpcli fail", logger.Any("err", err), xctx.CtxTraceIDField(ctx))
				return xerror.Wrap(err, "client do")
			}
			if response != nil {
				resp.Status = response.Status
				resp.StatusCode = response.StatusCode
				for k, v := range response.Header {
					resp.Header[k] = v
				}
				defer response.Body.Close()
				body, err := io.ReadAll(response.Body)
				if err != nil {
					return xerror.New(err.Error())
				}
				bodyBuf := bytes.NewBuffer(body)
				xctx.SetVal(ctx, fmt.Sprintf("oApiRsp-%s", xctx.CtxRequestID(ctx)), bodyBuf.Bytes())
				logger.Info("httpcli rsp", logger.Any("body", bodyBuf.String()), xctx.CtxTraceIDField(ctx))
				if body != nil {
					resp.Body = io.NopCloser(bodyBuf)
				}
			}
			return nil
		},
			retry.RetryTimes(req.retryCount),
			retry.RetryWithLinearBackoff(2*time.Second),
			retry.Context(ctx),
		)
	} else {
		response, e := client.Do(req.request)
		if e != nil {
			err = xerror.New(e.Error())
			if response != nil {
				resp.Status = response.Status
				resp.StatusCode = response.StatusCode
				for k, v := range response.Header {
					resp.Header[k] = v
				}
				resp.err = err
			}
			return resp, err
		}
		if response != nil {
			resp.Status = response.Status
			resp.StatusCode = response.StatusCode
			for k, v := range response.Header {
				resp.Header[k] = v
			}
			defer response.Body.Close()
			body, e := io.ReadAll(response.Body)
			if e != nil {
				err = xerror.New(e.Error())
			}
			bodyBuf := bytes.NewBuffer(body)
			xctx.SetVal(ctx, fmt.Sprintf("oApiRsp-%s", xctx.CtxRequestID(ctx)), bodyBuf.Bytes())
			logger.Info("httpcli rsp", logger.Any("body", bodyBuf.String()), xctx.CtxTraceIDField(ctx))
			if body != nil {
				resp.Body = io.NopCloser(bodyBuf)
			}
		}
	}

	if err != nil {
		logger.Error("httpcli fail", logger.Any("err", err), xctx.CtxTraceIDField(ctx))
	}

	req.response = resp
	req.err = err
	resp.err = err

	return resp, resp.err
}

// Response return response
func (req *Request) Response() (*Response, error) {
	if req.err != nil {
		return nil, req.err
	}
	return req.response, req.response.Error()
}

// -----------------------------------  Response -----------------------------------

// Error return err
func (resp *Response) Error() error {
	return resp.err
}

// BodyString returns the body data of the HttpResponse
func (resp *Response) BodyString() (string, error) {
	if resp.err != nil {
		return "", resp.err
	}
	body, err := resp.ReadBody()
	return string(body), err
}

// BodyString returns the body data of the HttpResponse
func (resp *Response) RspString() string {
	ret := make(map[string]any)
	ret["status"] = resp.Status
	ret["statusCode"] = resp.StatusCode
	ret["header"] = resp.Header
	if resp.err != nil {
		ret["err"] = resp.err.Error()
	}
	body, err := resp.ReadBody()
	if err == nil {
		ret["body"] = bytes.NewBuffer(body).String()
	}
	return xjson.ToJsonString(ret)
}

// ReadBody returns the body data of the HttpResponse
func (resp *Response) ReadBody() ([]byte, error) {
	if resp.err != nil {
		return []byte{}, resp.err
	}

	if resp.Body == nil {
		return []byte{}, errors.New("nil")
	}

	body, err := io.ReadAll(resp.Body)
	if err != nil {
		return []byte{}, err
	}
	resp.Body.Close()

	resp.Body = io.NopCloser(bytes.NewBuffer(body))
	return body, nil
}

// BindJSON parses the response's body as JSON
func (resp *Response) BindJSON(v interface{}) error {
	if resp.err != nil {
		return resp.err
	}
	body, err := resp.ReadBody()
	if err != nil {
		return err
	}
	return json.Unmarshal(body, v)
}

// -----------------------------------  Request way 2 -----------------------------------

// Option set options.
type Option func(*options)

type options struct {
	params  map[string]interface{}
	headers map[string]string
	timeout time.Duration
}

func (o *options) apply(opts ...Option) {
	for _, opt := range opts {
		opt(o)
	}
}

func defaultOptions() *options {
	return &options{}
}

// WithParams set params
func WithParams(params map[string]interface{}) Option {
	return func(o *options) {
		o.params = params
	}
}

// WithHeaders set headers
func WithHeaders(headers map[string]string) Option {
	return func(o *options) {
		o.headers = headers
	}
}

// WithTimeout set timeout
func WithTimeout(t time.Duration) Option {
	return func(o *options) {
		o.timeout = t
	}
}

// Get request, return custom json format
func Get(ctx context.Context, result interface{}, urlStr string, opts ...Option) error {
	o := defaultOptions()
	o.apply(opts...)
	return gDo(ctx, "GET", result, urlStr, o.params, o.headers, o.timeout)
}

// Delete request, return custom json format
func Delete(ctx context.Context, result interface{}, urlStr string, opts ...Option) error {
	o := defaultOptions()
	o.apply(opts...)
	return gDo(ctx, "DELETE", result, urlStr, o.params, o.headers, o.timeout)
}

// Post request, return custom json format
func Post(ctx context.Context, result interface{}, urlStr string, body interface{}, opts ...Option) error {
	o := defaultOptions()
	o.apply(opts...)
	return do(ctx, "POST", result, urlStr, body, o.params, o.headers, o.timeout)
}

// Put request, return custom json format
func Put(ctx context.Context, result interface{}, urlStr string, body interface{}, opts ...Option) error {
	o := defaultOptions()
	o.apply(opts...)
	return do(ctx, "PUT", result, urlStr, body, o.params, o.headers, o.timeout)
}

// Patch request, return custom json format
func Patch(ctx context.Context, result interface{}, urlStr string, body interface{}, opts ...Option) error {
	o := defaultOptions()
	o.apply(opts...)
	return do(ctx, "PATCH", result, urlStr, body, o.params, o.headers, o.timeout)
}

var requestErr = func(err error) error { return fmt.Errorf("request error, err=%v", err) }
var jsonParseErr = func(err error) error { return fmt.Errorf("json parsing error, err=%v", err) }
var notOKErr = func(resp *Response) error {
	body, err := resp.ReadBody()
	if err != nil {
		return err
	}
	if len(body) > 500 {
		body = append(body[:500], []byte(" ......")...)
	}
	return fmt.Errorf("statusCode=%d, body=%s", resp.StatusCode, body)
}

func do(ctx context.Context, method string, result interface{}, urlStr string, body interface{}, params KV, headers map[string]string, timeout time.Duration) error {
	if result == nil {
		return fmt.Errorf("'result' can not be nil")
	}

	req := &Request{}
	req.SetURL(urlStr)
	req.SetContentType("application/json")
	req.SetParams(params)
	req.SetHeaders(headers)
	req.SetBody(body)
	req.SetTimeout(timeout)

	var resp *Response
	var err error
	switch method {
	case "POST":
		resp, err = req.POST(ctx)
	case "PUT":
		resp, err = req.PUT(ctx)
	case "PATCH":
		resp, err = req.PATCH(ctx)
	}
	if err != nil {
		return requestErr(err)
	}
	defer resp.Body.Close() //nolint

	if resp.StatusCode != 200 {
		return notOKErr(resp)
	}

	err = resp.BindJSON(result)
	if err != nil {
		return jsonParseErr(err)
	}

	return nil
}

func gDo(ctx context.Context, method string, result interface{}, urlStr string, params KV, headers map[string]string, timeout time.Duration) error {
	req := &Request{}
	req.SetURL(urlStr)
	req.SetParams(params)
	req.SetHeaders(headers)
	req.SetTimeout(timeout)

	var resp *Response
	var err error
	switch method {
	case "GET":
		resp, err = req.GET(ctx)
	case "DELETE":
		resp, err = req.DELETE(ctx)
	}
	if err != nil {
		return requestErr(err)
	}
	defer resp.Body.Close() //nolint

	if resp.StatusCode != 200 {
		return notOKErr(resp)
	}

	err = resp.BindJSON(result)
	if err != nil {
		return jsonParseErr(err)
	}

	return nil
}

// StdResult standard return data
type StdResult struct {
	Code int         `json:"code"`
	Msg  string      `json:"msg"`
	Data interface{} `json:"data,omitempty"`
}

// StdResult standard return data
type StdResult2 struct {
	Code int         `json:"code"`
	Msg  string      `json:"message"`
	Data interface{} `json:"data,omitempty"`
}

// KV string:interface{}
type KV = map[string]interface{}
